import logging
import warnings

from .utils import fqn

logger = logging.getLogger('django_db_logging')

for hdlr in logger.handlers[:]:  # remove all old handlers
    if fqn(hdlr).startswith('django_db_logging.'):
        logger.removeHandler(hdlr)
        warnings.warn(UserWarning(
            f"Handler `{hdlr.name}` cannot be used with logger `django_db_logging`. "
            "It has been removed. Please fix your logging configuration."
        ), stacklevel=0)

if logger.propagate:  # pragma: no cover
    if logger.handlers:
        warnings.warn(UserWarning(
            f" `{logger.name}.propagate cannot` be `True`. "
            "It has been disabled. Please fix your logging configuration."
        ), stacklevel=0)
        logger.propagate = False



# Generated by Django 2.1.1 on 2018-09-25 09:03

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Record',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('timestamp', models.DateTimeField(auto_now_add=True, db_index=True)),
                ('logger', models.CharField(blank=True, max_length=200, null=True)),
                ('level', models.PositiveIntegerField(blank=True, choices=[(0, 'NOTSET'), (50, 'CRITICAL'), (40, 'ERROR'), (30, 'WARNING'), (20, 'INFO'), (10, 'DEBUG')], db_index=True, default=40)),
                ('message', models.TextField(blank=True, null=True)),
                ('formatted', models.TextField(blank=True, null=True)),
                ('server_name', models.CharField(db_index=True, max_length=128)),
                ('pathname', models.TextField(blank=True, null=True)),
                ('lineno', models.IntegerField(blank=True, null=True)),
                ('filename', models.TextField(blank=True, null=True)),
                ('module', models.TextField(blank=True, null=True)),
                ('func_name', models.TextField(blank=True, null=True)),
                ('process', models.TextField(blank=True, null=True)),
                ('traceback', models.TextField(blank=True, null=True)),
                ('exc_type', models.TextField(blank=True, null=True)),
                ('extra', models.TextField(blank=True, null=True)),
            ],
            options={
                'ordering': ('timestamp',),
                'get_latest_by': 'id',
            },
        ),
    ]



def LoggingRouter(alias="default"):
    class Router:

        def db_for_read(self, model, **hints):
            if model._meta.app_label == 'django_db_logging':
                return alias
            return None

        def db_for_write(self, model, **hints):
            if model._meta.app_label == 'django_db_logging':
                return alias
            return None

        def allow_migrate(self, db, app_label, model_name=None, **hints):
            if app_label == 'django_db_logging' and db == alias:
                return True

    return Router()



import logging
from logging import Handler

from django.apps import apps
from django.utils.functional import cached_property


class DBHandler(Handler):

    def __init__(self, model='django_db_logging.Record', level=logging.NOTSET):
        self.model_fqn = model
        super().__init__(level)

    @cached_property
    def model(self):
        m = apps.get_model(self.model_fqn)
        return m

    def emit(self, record):
        self.model.log_record(record, self.format(record))


class AsyncDBHandler(DBHandler):

    def __init__(self, model='django_db_logging.Record', level=logging.NOTSET):
        super().__init__(model, level)
        from .threadqueue import worker
        self._worker = worker

    def emit(self, record):
        self._worker.queue((self.model, record, self.format(record)))



import inspect


def get_classname(o):
    """ Returns the classname of an object r a class

    :param o:
    :return:
    """
    if inspect.isclass(o):
        target = o
    elif callable(o):
        target = o
    else:
        target = o.__class__
    try:
        return target.__qualname__
    except AttributeError:  # pragma: no cover
        return target.__name__


def fqn(o):
    """Returns the fully qualified class name of an object or a class

    :param o: object or class
    :return: class name

    >>> import django_db_logging
    >>> fqn('str')
    Traceback (most recent call last):
    ...
    ValueError: Invalid argument `str`
    >>> class A(object):
    ...     def method(self):
    ...         pass
    >>> str(fqn(A))
    'django_db_logging.utils.A'

    >>> str(fqn(A()))
    'django_db_logging.utils.A'

    >>> str(fqn(A.method))
    'django_db_logging.utils.A.method'

    >>> str(fqn(django_db_logging))
    'django_db_logging'


    """
    parts = []
    if hasattr(o, '__module__'):
        parts.append(o.__module__)
        parts.append(get_classname(o))
    elif inspect.ismodule(o):
        return o.__name__
    if not parts:
        raise ValueError("Invalid argument `%s`" % o)
    return ".".join(parts)



# -*- coding: utf-8 -*-
import datetime
import json
import logging
import os
import sys
import traceback
from logging import LogRecord

from django.db import models, connections
from django.db.transaction import atomic
from django_db_logging.handlers import DBHandler

from .logging import logger
from .settings import config

LOG_LEVELS = (
    (logging.NOTSET, logging._levelToName[logging.NOTSET]),
    (logging.CRITICAL, logging._levelToName[logging.CRITICAL]),
    (logging.ERROR, logging._levelToName[logging.ERROR]),
    (logging.WARNING, logging._levelToName[logging.WARNING]),
    (logging.INFO, logging._levelToName[logging.INFO]),
    (logging.DEBUG, logging._levelToName[logging.DEBUG]),)

PLAIN_RECORD = LogRecord("name", "level", "pathname", "lineno", "msg", None, None)

KNOWN_FIELDS = ("message",) + tuple(PLAIN_RECORD.__dict__.keys())


class RecordManager(models.Manager):
    def truncate(self):
        connection = connections[self.db]
        cursor = connection.cursor()
        cursor.execute('TRUNCATE TABLE "{0}"'.format(self.model._meta.db_table))

    def cleanup(self):
        with atomic():
            keep = self.all().order_by('-id').values_list('id', flat=True)[:config.MAX_LOG_ENTRIES]
            self.exclude(id__in=keep).delete()


class AbstractRecord(models.Model):
    timestamp = models.DateTimeField(auto_now_add=True, db_index=True)
    logger = models.CharField(max_length=200, blank=True, null=True, db_index=True)
    level = models.PositiveIntegerField(choices=LOG_LEVELS,
                                        default=logging.ERROR, blank=True, db_index=True)

    message = models.TextField(blank=True, null=True)
    formatted = models.TextField(blank=True, null=True)
    server_name = models.CharField(max_length=128, db_index=True)

    pathname = models.TextField(blank=True, null=True)
    lineno = models.IntegerField(blank=True, null=True)

    filename = models.TextField(blank=True, null=True)
    module = models.TextField(blank=True, null=True)
    func_name = models.TextField(blank=True, null=True)
    process = models.TextField(blank=True, null=True)

    traceback = models.TextField(blank=True, null=True)
    exc_type = models.TextField(blank=True, null=True)

    extra = models.TextField(blank=True, null=True)

    class Meta:
        ordering = ("timestamp", "id")
        get_latest_by = 'id'
        abstract = True

    objects = RecordManager()

    def __str__(self):
        return '<Record: %s, [%s] %s:%s, "%s">' % (self.logger, self.get_level_display(),
                                                   self.pathname, self.lineno, self.message)

    def extras(self):
        return json.loads(self.extra)

    @classmethod
    def log_record(cls, record: LogRecord, formatted: str = None):
        try:
            exc_info = {}
            extras = {k: str(v) for k, v in record.__dict__.items() if k not in KNOWN_FIELDS}

            if record.exc_info:
                exception, tb = record.exc_info[1:3]
                if not exception:
                    exc_type, exception, tb = sys.exc_info()
                else:
                    exc_type = exception.__class__
                if exc_type:
                    exc_info.update({'traceback': "\n".join(traceback.format_tb(tb)),
                                     'exc_type': exc_type.__name__})

                if config.USE_CRASHLOG:
                    try:
                        from crashlog.middleware import process_exception
                        process_exception(exception)
                    except ImportError:
                        pass

            cls.objects.create(
                timestamp=datetime.datetime.fromtimestamp(record.created),
                logger=record.name,
                level=record.levelno,
                message=record.getMessage(),
                formatted=formatted,
                pathname=record.pathname,
                filename=os.path.abspath(record.filename),
                module=record.module,
                func_name=record.funcName,
                lineno=record.lineno,
                process=record.process,
                extra=json.dumps(extras),
                **exc_info
            )
        except Exception as e:
            for hdlr in logger.handlers[:]:  # remove all old handlers
                if isinstance(hdlr, DBHandler):
                    logger.removeHandler(hdlr)
            logger.handle(record)
            logger.exception(e)


class Record(AbstractRecord):
    pass



#!/usr/bin/env python
import ast
import codecs
import os
import re
import subprocess
from distutils.errors import DistutilsError

from setuptools import find_packages, setup
from setuptools.command.sdist import sdist as BaseSDistCommand

ROOT = os.path.realpath(os.path.join(os.path.dirname(__file__)))
init = os.path.join(ROOT, 'src', 'django_db_logging', '__init__.py')

_version_re = re.compile(r'__version__\s+=\s+(.*)')
_name_re = re.compile(r'NAME\s+=\s+(.*)')

with open(init, 'rb') as f:
    content = f.read().decode('utf-8')
    version = str(ast.literal_eval(_version_re.search(content).group(1)))
    name = str(ast.literal_eval(_name_re.search(content).group(1)))

tests_requires = ['django-webtest',
                  'pytest',
                  'pytest-coverage',
                  'pytest-django',
                  'pytest-echo',
                  'pytest-pythonpath',
                  'tox',
                  ]
setup(
    name=name,
    version=version,
    url='https://github.com/saxix/django-db-logging',
    author='Stefano Apostolico',
    author_email='s.apostolico@gmail.com',
    license="MIT",
    description='Database logging handler with Django integration',
    package_dir={'': 'src'},
    packages=find_packages('src'),
    include_package_data=True,
    install_requires=["admin-extra-urls"],
    tests_require=tests_requires,
    extras_require={
        'test': tests_requires,
    },
    platforms=['linux'],
    classifiers=[
        'Environment :: Web Environment',
        'Framework :: Django',
        'Framework :: Django :: 1.11',
        'Framework :: Django :: 2.0',
        'Framework :: Django :: 2.1',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.6',
        'Intended Audience :: Developers'
    ]
)



import atexit
import os
import threading
from queue import Queue
from time import sleep, time

from django_db_logging.logging import logger


class ThreadQueue:
    _terminator = object()

    def __init__(self, shutdown_timeout=10):
        self._queue = Queue(-1)
        self._lock = threading.Lock()
        self._thread = None
        self._thread_for_pid = None
        self.options = {
            'shutdown_timeout': shutdown_timeout,
        }
        self.start()

    def is_alive(self):
        if self._thread_for_pid != os.getpid():
            return False
        return self._thread and self._thread.is_alive()

    def _ensure_thread(self):
        if self.is_alive():
            return
        self.start()

    def _timed_queue_join(self, timeout):
        """
        implementation of Queue.join which takes a 'timeout' argument

        returns true on success, false on timeout
        """
        deadline = time() + timeout
        queue = self._queue

        queue.all_tasks_done.acquire()
        try:
            while queue.unfinished_tasks:
                delay = deadline - time()
                if delay <= 0:
                    # timed out
                    return False

                queue.all_tasks_done.wait(timeout=delay)

            return True

        finally:
            queue.all_tasks_done.release()

    def main_thread_terminated(self):
        self._lock.acquire()
        try:
            if not self.is_alive():  # pragma: no cover
                # thread not started or already stopped - nothing to do
                return

            # wake the processing thread up
            self._queue.put_nowait(self._terminator)

            timeout = self.options['shutdown_timeout']

            # wait briefly, initially
            initial_timeout = 0.1
            if timeout < initial_timeout:
                initial_timeout = timeout

            if not self._timed_queue_join(initial_timeout):
                # if that didn't work, wait a bit longer
                # NB that size is an approximation, because other threads may
                # add or remove items
                size = self._queue.qsize()

                print(f"Logging sub-system is attempting to store {size} pending messages")
                print(f"Waiting up to {timeout} seconds")
                print("Press Ctrl-C to quit")

                self._timed_queue_join(timeout - initial_timeout)

            self._thread = None

        finally:
            self._lock.release()

    def _process(self, record):
        raise NotImplementedError

    def _target(self):
        while True:
            record = self._queue.get()
            try:
                if record is self._terminator:
                    break
                try:
                    self._process(record)
                except Exception:  # pragma: no cover
                    logger.error('Failed processing job', exc_info=True)
            finally:
                self._queue.task_done()

            sleep(0)

    def start(self):
        """
        Starts the task thread.
        """
        self._lock.acquire()
        try:
            if not self.is_alive():
                self._thread = threading.Thread(target=self._target)
                self._thread.setDaemon(True)
                self._thread.start()
                self._thread_for_pid = os.getpid()
        finally:
            self._lock.release()
            atexit.register(self.main_thread_terminated)

    def queue(self, payload):
        self._ensure_thread()
        self._queue.put_nowait(payload)

    def stop(self, timeout=None):
        """
        Stops the task thread. Synchronous!
        """
        self._lock.acquire()
        try:
            if self._thread:
                self._queue.put_nowait(self._terminator)
                self._thread.join(timeout=timeout)
                self._thread = None
                self._thread_for_pid = None
        finally:
            self._lock.release()


class ThreadLogger(ThreadQueue):
    def _process(self, record):
        model, log_record, formatted = record
        model.log_record(log_record, formatted)


worker = ThreadLogger()
worker.start()



import logging

from django.conf import settings

DEFAULTS = {
    "MAX_LOG_ENTRIES": 10000,
    "LEVELS": [logging.CRITICAL,
               logging.ERROR,
               logging.WARNING,
               logging.INFO,
               logging.DEBUG],
    "ENABLE_TEST_BUTTON": True,
    "USE_CRASHLOG": False,
    "ALLOW_TRUNCATE": True,
    # "DATABASE": "default"
}


class Config:
    pass


config = Config()

custom = getattr(settings, 'DBLOG', {})

for key, value in DEFAULTS.items():
    custom.setdefault(key, value)
    setattr(config, key, custom[key])



# -*- coding: utf-8 -*-
import logging

from django.contrib import admin
from django.contrib.admin import ModelAdmin

from admin_extra_urls.extras import ExtraUrlMixin, link
from admin_extra_urls.mixins import _confirm_action
from django.contrib.admin.views.main import ChangeList
from django.core.paginator import Paginator, InvalidPage
from django.utils.html import format_html

from django_db_logging.handlers import DBHandler
from django_db_logging.settings import config
from .models import Record


class PlainPaginator(Paginator):
    def _get_count(self):
        return config.MAX_LOG_ENTRIES  # we never have more than this
    count = property(_get_count)


class CustomChangeList(ChangeList):
    def get_results(self, request):
        paginator = PlainPaginator(self.queryset, self.list_per_page)
        try:
            result_list = paginator.page(self.page_num + 1).object_list
        except InvalidPage:  # pragma: no cover
            result_list = ()

        self.full_result_count = "~"
        self.result_count = "~"
        self.result_list = result_list
        self.can_show_all = False
        self.multi_page = True
        self.paginator = paginator


@admin.register(Record)
class RecordAdmin(ExtraUrlMixin, ModelAdmin):
    list_display = ('timestamp', 'lvl', 'logger', 'message', 'exception')
    list_filter = ('timestamp', 'level',)
    search_fields = ('logger', )
    date_hierarchy = 'timestamp'
    change_list_template = 'django_db_logging/change_list.html'
    change_form_template = 'django_db_logging/change_form.html'

    def has_add_permission(self, request):
        return False

    def get_changelist(self, request, **kwargs):
        return CustomChangeList

    def lvl(self, instance):
        level = instance.get_level_display()
        return format_html(f'<span class="badge {level}">{level}</span>')
    lvl.short_description = 'Level'

    def exception(self, instance):
        return bool(instance.exc_type)
    exception.short_description = 'Exception'
    exception.boolean = True


    @link(label='Cleanup')
    def cleanup(self, request):
        self.model.objects.cleanup()
        self.message_user(request, "Log cleaned")

    @link(label='Empty Log', css_class="btn btn-danger",
          permission=lambda *a: config.ALLOW_TRUNCATE,
          icon="icon-trash icon-white")
    def empty_log(self, request):
        def _action(request):
            self.model.objects.truncate()

        return _confirm_action(self, request, _action,
                               "Confirm deletion whole error log",
                               "Successfully executed",
                               template='django_db_logging/confirm.html', )

    @link(permission=lambda *a: config.ENABLE_TEST_BUTTON)
    def test(self, request):
        logger = logging.getLogger('django_db_logging_test_logger')
        logger.propagate = False

        handler = DBHandler()
        formatter = logging.Formatter('%(extra_param)s : %(message)s')
        handler.setFormatter(formatter)
        logger.addHandler(handler)

        extra = {'extra_param': 'EXTRA_PARAM'}
        logger.info("DBLogger [INFO] test log entry", extra=extra)
        logger.debug("DBLogger [DEBUG] test log entry", extra=extra)
        logger.error("DBLogger [ERROR] test log entry", extra=extra)
        logger.warning("DBLogger [WARNING] test log entry", extra=extra)
        logger.critical("DBLogger [CRITICAL] test log entry", extra=extra)
        logger.info("DBLogger [INFO] test log entry", extra=extra)

        try:
            raise Exception("DBLogger [EXCEPTION] test log entry")
        except Exception as e:
            logger.exception(e, extra=extra)
